/*
 * Copyright (c) 2004 Nokia
 * Author: Timo Savola <tsavola@movial.fi>
 *
 * Licensed under GPL, see COPYING for details.
 */

#include <map>
#include <set>
#include <list>
#include <string>
#include <fstream>
#include <iostream>
#include <cctype>
#include <sys/types.h>
#include <dirent.h>

using namespace std;

#define DEVKIT_ROOT "/scratchbox/devkits/"
#define DEVKIT_DEPS "/depends"

int main(int argc, char **argv)
{
	list<string> names;
	map<string, set<string> > depmap;

	/*
	 * Get a devkit listing and a map of their dependencies.
	 */

	if (DIR *dir = opendir(DEVKIT_ROOT)) {
		while (struct dirent *ent = readdir(dir)) {
			if (ent->d_name[0] == '.')
				continue;

			string path = DEVKIT_ROOT;
			path += ent->d_name;
			path += DEVKIT_DEPS;

			set<string> deps;

			ifstream file(path.c_str());
			if (file.is_open()) {
				string name;

				while (true) {
					int c = file.get();

					if (file.fail()) {
						if (!name.empty())
							deps.insert(name);
						break;
					}

					if (!isspace(c) && c != ':') {
						name += (char) c;
					} else if (!name.empty()) {
						deps.insert(name);
						name.clear();
					}
				}
			}

			names.push_back(ent->d_name);
			depmap[ent->d_name] = deps;
		}

		closedir(dir);
	}

	/*
	 * Reorder devkits so that the dependencies of a devkit precede it.
	 * Implemented by moving the devkit forward in the list if necessary.
	 */

	set<string> done;

	for (list<string>::iterator i = names.begin(); i != names.end(); ) {
		const set<string> &deps = depmap[*i];

		unsigned int count = deps.size();
		bool found = false;
		list<string>::iterator new_pos;

		if (done.insert(*i).second && count > 0)
			for (list<string>::iterator j = names.begin(); j != names.end(); ++j)
				if (*j == *i) {
					found = true;
				} else if (deps.find(*j) != deps.end() && --count == 0) {
					new_pos = ++j;
					break;
				}

		if (found) {
			names.insert(new_pos, *i);
			names.erase(i++);
		} else {
			++i;
		}
	}

	/*
	 * Determine output mode.
	 */

	if (argc < 2) {
		cerr << "Usage: " << argv[0] << " deps   [devkit names]" << endl
		     << "       " << argv[0] << " expand <devkit names>" << endl;
		return 1;
	}

	string mode = argv[1];

	/*
	 * Determine names of devkits which to print.
	 */

	set<string> filter;
	list<string> nonexistent;

	for (int i = 2; i < argc; ++i) {
		string s = argv[i];
		if (done.find(s) == done.end())
			nonexistent.push_back(s);
		else if (s != "none")
			filter.insert(argv[i]);
	}

	if (!nonexistent.empty()) {
		cerr << "No such devkit(s):";
		for (list<string>::const_iterator i = nonexistent.begin(); i != nonexistent.end(); ++i)
			cerr << ' ' << *i;
		cerr << endl;
		return 1;
	}

	/*
	 * Print the names of devkits mapped to their dependencies using line format:
	 *   <devkit>: <dep1> <dep2> ... <depN>
	 */

	if (mode == "deps") {
		for (list<string>::const_iterator i = names.begin(); i != names.end(); ++i) {
			if (!filter.empty() && filter.find(*i) == filter.end())
				continue;

			cout << *i << ':';

			const set<string> &deps = depmap[*i];
			for (set<string>::const_iterator j = deps.begin(); j != deps.end(); ++j)
				cout << ' ' << *j;

			cout << endl;
		}

	/*
	 * Print the names of devkits including their dependencies in a single list.
	 */

	} else if (mode == "expand") {
		set<string> full = filter;

		bool changed;
		do {
			changed = false;

			for (set<string>::const_iterator i = full.begin(); i != full.end(); ++i) {
				const set<string> &deps = depmap[*i];
				for (set<string>::const_iterator j = deps.begin(); j != deps.end(); ++j)
					if (full.insert(*j).second)
						changed = true;
			}

		} while (changed);

		for (list<string>::const_iterator i = names.begin(); i != names.end(); ++i)
			if (full.find(*i) != full.end())
				cout << *i << endl;

	} else {
		cerr << "sb_devkits: unknown output mode: " << mode << endl;
		return 1;
	}

	return 0;
}
